{Linked list unit By Alan Birtles}

unit linklists;

interface
uses sysutils,classes,rtlconsts;

//data type used when adding to a list
type TAfterBefore=(LAfter,LBefore);
     TCompareResult=(CrLess,CrEqual,CrMore);

     //function skeleton to compare two pointers
     TCompare=function (const data1,data2:pointer):TCompareResult;

     //pointer to a linked list cell
     PLinkListCell=^TLinkListCell;
     //linked list cell type
     TLinkListCell =record
               prev,next:PLinkListCell;
               Data:pointer;
               end;
     //marker for faster list traversal
     TMarker =record
             Cell:PLinkListCell;
             Number:integer;
             end;

     //generic linked list type
     TLinkList=class(TObject)
     protected
        FFirst,FLast,FCurrentCell:PLinkListCell;
        FLength,FCurrentno:integer;
        FSorted:Boolean;
        flastsort:TCompare;
        FLastSortRev:boolean;
        fonchange:TNotifyEvent;
        Markers:array of tmarker;
        FStart:integer;
        function getcurrentno:integer;
        function fgetcurrent:pointer;
        function fgetlast:pointer;
        function fgetfirst:pointer;
        property first:pointer read fgetfirst;
        property last:pointer read fgetlast;
        property CurrentCell:pointer read fgetcurrent;
        property Length:integer read FLength;
        property CurrentNo:integer read getcurrentno;
        property sorted:boolean read FSorted;
        property StartNo:integer read FStart;
        procedure addBeg(const data:pointer);
        procedure add(const data:pointer;const before:TAfterBefore);overload;
        procedure add(const data:pointer;const position:integer;const before:TAfterBefore);overload;
        procedure addend(const data:pointer);
        procedure DeleteAll(const KeepData:Boolean);dynamic;
        procedure Delete(const i:integer);overload;dynamic;
        procedure Delete(const Keepdata:boolean;const j:integer);overload;dynamic;
        procedure Delete;overload;dynamic;
        function GetNo(const no:integer):pointer;dynamic;
        function getnext:pointer;
        function getprev:pointer;
        function getfirst:pointer;
        function getlast:pointer;
        procedure quicksort(const by:TCompare;const reverse:boolean);
        procedure CreateMarkers(const interval:integer);
        procedure ReplaceCurrent(const data:pointer);
        function binsearch(const data:pointer):tcompareresult;
        function Search(const data:pointer;const by:TCompare):boolean;
        procedure insert(const data:pointer);
        procedure randomize;
        procedure swap(const a,b:integer);
        property onchange:TNotifyEvent read fonchange write fonchange;
        function lastsort(const data1,data2:pointer):TCompareResult;
     public
        constructor create;overload;
        constructor create(const StartingNumber:integer);overload;
        destructor destroy;override;
     end;

procedure copy(var fromlist,tolist:TLinklist);overload;
procedure copy(var fromlist:tlinklist;const pos:integer;var tolist:TLinklist);overload;
procedure invcomp(var comp:TCompareResult);
function strcomp(const p1,p2:string):tcompareresult;

implementation

procedure copy(var fromlist,tolist:TLinklist);
//copy the current cell's data from fromlist to to list
//doesn't create new memory
begin
tolist.insert(fromlist.CurrentCell);
end;

procedure copy(var fromlist:tlinklist;const pos:integer;var tolist:TLinklist);
//copy data from pos in fromlist to tolist
begin
fromlist.GetNo(pos);
copy(fromlist,tolist);
end;

procedure invcomp(var comp:TCompareResult);
//invert a compare result
begin
case comp of
     crless:comp:=crmore;
     crmore:comp:=crless;
     end;
end;

function strcomp(const p1,p2:string):tcompareresult;
//compare two pointers to strings alphabetically
var r:integer;
begin
r:=sysutils.stricomp(pchar(p1),pchar(p2));
result:=crequal;
if r<0 then result:=crless;
if r>0 then result:=crmore;
end;

function tlinklist.lastsort(const data1,data2:pointer):TCompareResult;
begin
if not assigned(flastsort) then raise EListError.Create('List is unsorted');
result:=flastsort(data1,data2);
if flastsortrev then invcomp(result);
end;

function tlinklist.fgetcurrent:pointer;
//return the current cell's data
begin
if fcurrentcell<>nil then result:=fcurrentcell.Data
   else result:=nil;
end;

function tlinklist.fgetlast;
//get the last cell in the list's data
begin
if flast<>nil then result:=flast.Data
   else result:=nil;
end;

function tlinklist.fgetfirst;
//get the first cell in the list's data
begin
if ffirst<>nil then result:=ffirst.Data
   else result:=nil;
end;

procedure tlinklist.randomize;
//randomize a list
var i:integer;
begin
system.randomize;
for i:=fstart to fstart+flength-1 do
    swap(i,random(fstart+flength));
fsorted:=false;
end;

procedure tlinklist.swap;
//swap two items in a list
var pa,pb:Plinklistcell;
    tmp:pointer;
begin
getno(a);
pa:=currentcell;
getno(b);
pb:=currentcell;
tmp:=pa.Data;
pa.Data:=pb.Data;
pb.Data:=tmp;
end;

procedure tlinklist.DeleteAll;
//delete all items from the list if keepdata then the data is not freed
var tmp:plinklistcell;
begin
while ffirst<>nil do begin
      tmp:=ffirst;
      Ffirst:=Ffirst^.next;
      if (not keepdata)and(tmp^.data<>nil) then dispose(tmp^.Data);
      dispose(tmp);
      end;
//set the list to empty
flast:=nil;
fcurrentcell:=nil;
fcurrentno:=0;
flength:=0;
if assigned(fonchange) then fonchange(self);
end;

procedure tlinklist.AddBeg(const data:pointer);
//add item to begining of the list
begin
getfirst;
add(data,LBefore);
fsorted:=false;
end;

procedure tlinklist.Add(const data:pointer;const Before:TAfterBefore);
//add item before/after current cell
var tmp:Plinklistcell;
    i:integer;
begin
if Flength=0 then begin
   //list is empty add item
   new(Ffirst);
   Ffirst.prev:=nil;
   Ffirst.next:=nil;
   Ffirst.Data:=data;
   Flast:=Ffirst;
   Fcurrentcell:=Ffirst;
   Flength:=1;
   Fcurrentno:=1;
   end else begin
       new(tmp);
       case before of
            lafter: begin
                   //add item after the current cell
                   tmp^.prev:=Fcurrentcell;
                   tmp^.next:=Fcurrentcell^.next;
                   if Fcurrentcell^.next<>nil then Fcurrentcell^.next^.prev:=tmp;
                   Fcurrentcell^.next:=tmp;
                   tmp^.data:=data;
                   if Fcurrentcell=Flast then Flast:=tmp;
                   Fcurrentcell:=tmp;
                   inc(Flength);
                   inc(fcurrentno);
                   for i :=low(markers) to high(markers) do begin
                       if markers[i].number>fcurrentno then inc(markers[i].number);
                       end;
                   end;
            lbefore: begin
                   //add item before the current cell
                   tmp^.next:=fcurrentcell;
                   tmp^.prev:=fcurrentcell^.prev;
                   if fcurrentcell^.prev<>nil then fcurrentcell^.prev^.next:=tmp;
                   fcurrentcell^.prev:=tmp;
                   tmp^.data:=data;
                   if fcurrentcell=ffirst then Ffirst:=tmp;
                   Fcurrentcell:=tmp;
                   inc(Flength);
                   for i :=low(markers) to high(markers) do begin
                       if markers[i].number>=fcurrentno then inc(markers[i].Number);
                       end;
                   end;
            end;
       end;
fsorted:=false;
if assigned(fonchange) then fonchange(self);
end;

procedure tlinklist.Add(const data:pointer;const Position:integer;const before:TAfterBefore);
//add item before/after number
begin
getno(position);
add(data,before);
fsorted:=false;
end;

procedure tlinklist.AddEnd(const data:pointer);
//add item to end of list
begin
getlast;
add(data,lafter);
fsorted:=false;
end;

procedure tlinklist.Delete(const i:integer);
//delete item number don't kepp data
begin
delete(false,i);
end;

procedure tlinklist.Delete(const keepdata:boolean;const j:integer);
//delete item keeping data if keepdata is true
var tmp:PLinkListcell;
    i:integer;
begin
getno(j);
if (flength=0) or (fcurrentcell=nil)then exit;
if flast=fcurrentcell then if fcurrentcell<>nil then flast:=fcurrentcell^.prev;
if ffirst=fcurrentcell then begin
   //first item in list
   if ffirst^.next<>nil then ffirst^.next^.Prev:=nil;
   tmp:=ffirst;
   Ffirst:=Ffirst^.Next;
   dec(Flength);
   Fcurrentcell:=Ffirst;
   if (system.length(markers)<>0) and (markers[0].Number=0) then markers[0].Cell:=tmp^.next;
   if not keepdata then dispose(tmp^.Data);
   dispose(tmp);
   for i := low(markers) to high(markers) do
       if markers[i].Number>0 then dec(markers[i].Number);
   end else begin
       if fcurrentcell^.next<>nil then begin
          fcurrentcell^.next^.Prev:=fcurrentcell^.Prev;
          fcurrentcell^.Prev^.Next:=fcurrentcell^.next;
          end else begin
              fcurrentcell^.Prev^.Next:=nil;
              end;
          tmp:=fcurrentcell;
          Fcurrentcell:=Fcurrentcell^.Prev;
          for i :=low(markers) to high(markers) do
              if markers[i].Number=fcurrentno then
                 markers[i].Cell:=markers[i].Cell^.next else
                 if markers[i].number>fcurrentno then dec(markers[i].number);
          dec(Fcurrentno);
          dec(Flength);
          if not keepdata then dispose(tmp^.Data);
          dispose(tmp);
          end;
if assigned(fonchange) then fonchange(self);
end;

procedure tlinklist.Delete;
//delete current item and data
begin
delete(false,currentno);
end;

function tlinklist.GetNo(const no:integer):pointer;
//goto number
var closestmarker,i,no1:integer;
begin
no1:=no-fstart+1;
if no1=fcurrentno then begin
   //list already in correct position
   result:=currentcell;
   exit;
   end;
if (no1>flength)or(no1<=0) then raise EListError.CreateResFmt(@SListIndexError, [no1]);
closestmarker:=0;
//if markers exist find closest one to desired item
if system.length(markers)>1 then
   for i:= 1 to high(markers) do
       if ((markers[i].number>0)and(abs(markers[i].Number-no1)<abs(markers[closestmarker].Number-no1))) or (markers[closestmarker].Number<1) then
          closestmarker:=i;
//find best position for list (beggining, end or current)
if (abs(no1-fcurrentno)<no1)
   then if(abs(no1-fcurrentno)>length-no1) then begin
        //end of list is best positon
        fcurrentno:=length;
        fcurrentcell:=flast;
        end else//list is already in best position
   else begin
        //start of list is best position
        fcurrentno:=1;
        fcurrentcell:=ffirst;
        end;
//determine whether current position or best marker is best position
if (system.length(markers)>0)and(markers[closestmarker].Number>0) then begin
   if abs(fcurrentno-no1)>abs(markers[closestmarker].number-no1) then begin
      fcurrentno:=markers[closestmarker].Number;
      fcurrentcell:=markers[closestmarker].Cell;
      end;
   end;
//move up or down list from current position to desired position
if fcurrentno<no1 then begin
   while fcurrentno<>no1 do
         getnext;
   end else begin
       while fcurrentno<>no1 do
             getprev;
       end;
result:=currentcell;
if assigned(fonchange) then fonchange(self);
end;

constructor tlinklist.create;
//create a list setting the starting item to 1 change this by calling create(no)
begin
create(1);
end;

constructor tlinklist.create(const StartingNumber:integer);
//create an empty list and set the starting number
begin
inherited create;
Ffirst:=nil;
Flast:=nil;
Fcurrentcell:=nil;
fstart:=startingNumber;
Flength:=0;
Fcurrentno:=0;
fsorted:=false;
setlength(markers,0);
end;


destructor tlinklist.destroy;
//destroy the linklist if you have data you want to keep call deleteall(true) first
begin
deleteall(false);
inherited destroy;
end;

function tlinklist.getnext;
//get next cell
begin
if Fcurrentcell<>Flast then begin
   Fcurrentcell:=Fcurrentcell.next;
   inc(Fcurrentno);
   end else raise EListError.CreateResFmt(@SListIndexError, [flength]);
result:=currentcell;
if assigned(fonchange) then fonchange(self);
end;

function tlinklist.getprev;
//get previous cell
begin
if Fcurrentcell<>Ffirst then begin
   Fcurrentcell:=Fcurrentcell.prev;
   dec(Fcurrentno);
   end else raise EListError.CreateResFmt(@SListIndexError, [0]);
result:=currentcell;
if assigned(fonchange) then fonchange(self);
end;

function tlinklist.getfirst;
//get first cell
begin
Fcurrentcell:=ffirst;
fcurrentno:=1;
result:=currentcell;
if assigned(fonchange) then fonchange(self);
end;

function tlinklist.getlast;
//get last cell
begin
fcurrentcell:=flast;
fcurrentno:=flength;
result:=currentcell;
if assigned(fonchange) then fonchange(self);
end;

procedure tlinklist.quicksort;
//sort ascending order if "reverse" is false and
//"by" returns less when d1 is less than d2

procedure QSplit(list:PLinkListcell;var l1,l2:PLinkListCell;const By:Tcompare;
                 const reverse:Boolean);
//split a list into two sublists of items more than first item in list
//and items less than first item in list
var tmp:plinklistcell;
    top:Plinklistcell;
    less:tcompareresult;
begin
top:=list;
list:=list.next;
while list<>nil do begin
      less:=by(top^.data,list^.data);
      if not reverse then invcomp(less);
      if less=crless then begin
         if l1=nil then begin
            //set l1 to item
            new(l1);
            l1.Next:=nil;
            l1.Prev:=nil;
            l1.data:=list.data;
            end else begin
                //add item to l1
                tmp:=l1;
                new(l1);
                l1.Next:=tmp;
                l1.Prev:=nil;
                tmp.Prev:=l1;
                l1.data:=list.data;
                end;
         end
         else begin
              if l2=nil then begin
                 //set l2 to item
                 new(l2);
                 l2.Next:=nil;
                 l2.Prev:=nil;
                 l2.data:=list.data;
                 end else begin
                     //add item to l2
                     tmp:=l2;
                     new(l2);
                     l2.Next:=tmp;
                     l2.Prev:=nil;
                     tmp.Prev:=l2;
                     l2.data:=list.data;
                     end;
              end;
      tmp:=list;
      list:=list.Next;
      dispose(tmp);
      end;
//set start of list's previous to nil
if l1<>nil then l1.Prev:=nil;
if l2<>nil then l2.Prev:=nil;
end;

procedure getlistlast(var list,last:plinklistcell);
//get end of list
begin
last:=list;
while (last<>nil)and(last.next<>nil) do last:=last^.next;
end;

Procedure QSort(var list:PLinkListCell;const By:Tcompare;const reverse:Boolean);
//recursively quick sort list
var l1,l2,tmp:PLinkListCell;
begin
if list=nil then exit;//no elements so sorted
if list.next=nil then exit;//one element so sorted
l1:=nil;
l2:=nil;
setlength(markers,0);
//split list
qsplit(list,l1,l2,By,reverse);
//sort sublists
if (l1<> nil)and(l1.Next<>nil) then qsort(l1,by,reverse);
if (l2<>nil)and(l2.next<>nil) then qsort(l2,by,reverse);
//combine sublists
if l2=nil then begin
   new(l2);
   l2^.data:=list^.data;
   l2.Next:=nil;
   l2.Prev:=nil;
   dispose(list);
   end else begin
       tmp:=l2;
       new(l2);
       l2.Prev:=nil;
       l2.Next:=tmp;
       tmp.Prev:=l2;
       l2^.data:=list^.data;
       dispose(list);
       end;
if l1<>nil then begin
   list:=l1;
   getlistlast(list,l1);
   if l1<> nil then l1.Next:=l2;
   if l2<> nil then l2.Prev:=l1;
   end else list:=l2;
end;

begin
if length<>0 then begin
   //sort list
   qsort(ffirst,by,reverse);
   //reset list porperties
   fcurrentcell:=ffirst;
   while (fcurrentcell.next<>nil) and (fcurrentcell<>nil) do begin
         fcurrentcell:=fcurrentcell.next;
         end;
   flast:=fcurrentcell;
   fcurrentcell:=ffirst;
   fcurrentno:=1;
   end;
//store parameters used to sort for later inserts and binary searches
fsorted:=true;
flastsort:=by;
flastsortrev:=reverse;
//delete markers
setlength(markers,0);
if assigned(fonchange) then fonchange(self);
end;

function tlinklist.Search;
//search the list for a value.
//binsearch is much fatser if list is correctly sorted
begin
getfirst;
result:=false;
while (currentcell<>last) and (result=false)do
      if by(data,currentcell)<>crequal then getnext else result:=true;
//check last element
if (result=false)and(by(data,currentcell)=crequal)then result:=true;
if assigned(fonchange) then fonchange(self);
end;

function tlinklist.binsearch;
//perform a binary search. List MUST be sorted prior to calling this
//set current cell to the closest cell.
//result determines whether the current cell is more less or equal to
//the serached for item
var first,last,middle:integer;
    fm,lm,mm:integer;
    comp:tcompareresult;
begin
if not fsorted then begin
   raise elisterror.Create('Cannot Binary Search Unsorted List');
   exit;
   end;
if length=0 then begin
   result:=crless;
   exit;
   end;
result:=crless;
first:=1;
last:=flength;
getfirst;
//check first item is not equal
if lastsort(data,currentcell)=crequal then begin
   result:=crequal;
   end else begin
   //check last item is not equal
   getlast;
   if lastsort(data,currentcell)=crequal then result:=crequal;
   end;
//check first item is not less
if (result<>crequal) then begin
   getfirst;
   comp:=lastsort(data,ffirst.data);
   if comp=crless then begin
      result:=crless;
      exit;
      end;
   end;
//check last item is not more
if (result<>crequal) then begin
   getlast;
   comp:=lastsort(data,flast.data);
   if comp=crmore then begin
      result:=crmore;
      exit;
      end;
   end;
//check markers
if system.Length(markers)<>0 then begin
   fm:=0;
   lm:=system.length(markers)-1;
   mm:=0;
   //check first marker
   comp:=lastsort(data,markers[fm].Cell.data);
   if comp= crequal then begin
      fcurrentcell:=markers[fm].Cell;
      fcurrentno:=markers[fm].Number;
      result:=crequal;
      exit;
      end else begin
          //check before first marker
          if comp=crless then begin
             mm:=-1;
             last:=markers[fm].Number;
             end;
          end;
   //check last marker
   comp:=lastsort(data,markers[lm].Cell.data);
   if comp=crequal then begin
      fcurrentcell:=markers[lm].Cell;
      fcurrentno:=markers[lm].Number;
      result:=crequal;
      exit;
      end else begin
          //check after last marker
          if comp=crmore then begin
             mm:=-1;
             first:=markers[lm].Number;
             end;
          end;
   while (result<>crequal)and(fm<=lm)and(mm<>-1) do begin
         mm:=(fm+lm)shr 1;
         result:=lastsort(data,markers[mm].Cell.Data);
         case result of
              crequal:begin
                      fcurrentcell:=markers[mm].Cell;
                      fcurrentno:=markers[mm].Number;
                      result:=crequal;
                      end;
              crmore: fm:=mm+1;
              crless: lm:=mm-1;
              end;
         end;
   if mm<>-1 then begin
      //found a marker to start from
      case result of
           crequal:;//do nothing
           crmore: begin //data is more than markers[mm]
                   first:=markers[mm].Number;
                   if mm<system.Length(markers)-1 then last:=markers[mm+1].Number;
                   end;
           crless: begin //data is less than markers[mm]
                   last:=markers[mm].Number;
                   if mm>1 then first:=markers[mm-1].Number;
                   end;
           end;
      end;
   end;
while (result<>crequal)and(first<=last) do begin
      middle:=(first+last)shr 1;
      getno(middle);
      comp:=lastsort(data,currentcell);
      result:=comp;
      case comp of
           crequal:{result:=crequal};//already equal
           crmore: begin
                   //if go up
                   first:=middle+1;
                   end;
           crless: begin
                   //if go down
                   last:=middle-1;
                   end;
           end;
      end;
end;

procedure tlinklist.insert;
//insert an item into a sorted list
//puts at start if list is not sorted
var sort:boolean;
    comp:tcompareresult;
begin
sort:=fsorted;
if length=0 then begin
   addbeg(data);
   end else
       if fsorted then begin
          comp:=binsearch(data);
          case comp of
               crless,crequal:add(data,lbefore);
               else add(data,lafter);
               end;
          end else add(data,lbefore);
fsorted:=sort;
end;

procedure tlinklist.createmarkers;
//Create markers in the list to improve binserach and goto performance
//an interval of 50-100 in a large list is about optimum
//dont bother if list is less than 200 or so elements
//update markers regularly to make them worthwhile
//markers are deleted after major operations such as sorting
begin
setlength(markers,0);
getfirst;
while length-fcurrentno>interval+1 do begin
      while fcurrentno mod interval<>0 do begin
            getnext;
            end;
      getnext;
      setlength(markers,system.length(markers)+1);
      markers[high(markers)].Cell:=fcurrentcell;
      markers[high(markers)].Number:=fcurrentno;
      end;
if assigned(fonchange) then fonchange(self);
end;

function tlinklist.getcurrentno;
//returns the number of the current cell
begin
result:=fstart+fcurrentno-1;
end;

procedure tlinklist.ReplaceCurrent;
//replaces currentcell's data with data
begin
if (length=0) then raise EListError.Create('List is Empty, cannot replace data');
if (fcurrentcell.data<>nil) then dispose(fcurrentcell.Data);
fcurrentcell.Data:=data;
end;

end.

